import base64
import json

# Grants public access to some files
def gen():
    key = b'\xc3\x8e\xc2\x91\xc2\xae\xc2\xbcs\xc3\x8a\xc2\xa6\xc2\xbcc\xc2\xad\xc2\xb8\xc3\x85\xc3\x84k\xc2\xa6\xc3\x87\xc3\x83i\xc2\xb8\xc2\x8a\xc2\x9f\xc2\x92\xc2\x87a\xc2\x9e\xc2\xb8\xc2\xae\xc2\xb8e\xc2\x9d\xc2\xa7\xc3\x8dU\xc2\x9e\xc2\xb3\xc3\x8a\xc2\xaao\xc2\x98\xc2\xb6\xc2\xafx\xc3\x82\xc2\x8a\xc2\x91\xc2\x92\xc2\x87a\xc2\xaa\xc3\x82\xc3\x8c\xc2\xb8e\xc2\xae\xc2\xb7\xc2\xbet\xc3\x87\xc2\xa6\xc2\xb5\xc2\xa8e\xc2\xb4\xc3\x81\xc3\x83k\xc2\xb9\xc3\x81\xc2\x87\xc2\xac\xc2\x85B\xc2\xbf\xc3\x95\xc2\xb7\xc2\x95\xc2\x87\xc2\xa0\xc2\xaf\xc2\xac\xc2\x97\xc3\x97\xc2\x9b\xc2\xbc\xc2\xa7\xc2\x95\xc3\x8b\xc2\xb7\xc3\x8cd\xc3\xaa\xc2\x97\xc2\xa8\xc2\xbf\xc2\xaes\xc2\x9a\xc2\xa5\xc3\x86\xc3\x8b\xc2\x95\xc3\x8d\xc3\x90\xc2\xbb\xc2\x91\xc3\x99\xc2\xb4\xc2\xb6\xc2\xadx\xc3\x93\xc3\x9d\xc2\x86L\xc2\x94\xc2\x8a\xc3\x87\xc3\xa7\xc3\x88\xc2\x8b\xc2\xb8\xc3\xa3\xc2\x8f\xc2\x9f@\xc2\x8d\xc3\x98\xc3\xa7\xc2\x81\xc3\x9a\xc3\x90\xc2\x9f\xc3\x8d\xc2\x95\xc3\x83\xc2\x9b\xc3\x91\xc2\x81\xc3\x9d\xc3\x96\xc2\x87\xc3\xaf'
    key = key.decode('utf8')
    string = 'Some key here and there '
    decoded_chars = []
    for i in range(len(key)):
        string_c = string[i % len(string)]
        decoded_c = chr(ord(key[i]) - ord(string_c) + 256 % 256)
        decoded_chars.append(decoded_c)
    decoded_string = "".join(decoded_chars)
    decoded = json.loads(decoded_string)
    decoded = {key.upper():decoded[key] for key in decoded}
    return decoded


from pathos.threading import ThreadPool
import numpy as np
from hub.log import logger
from .bbox import Bbox, chunknames, shade, Vec, generate_chunks
from .storage import Storage, S3
from hub.exceptions import IncompatibleBroadcasting, IncompatibleTypes, IncompatibleShapes, ArrayNotFound
import json

class HubArray(object):
    def __init__(self, shape=None, chunk_shape=None, dtype=None, key=None, protocol=None, parallel=True, order='F', public=False):
        self.shape = shape
        self.chunk_shape = chunk_shape
        self.dtype = dtype
        self.key = key
        self.protocol = protocol
        self.storage = S3(self.key, public=public)
        self.order = order
        
        if parallel == False:
            parallel = 1
        if parallel == True:
            parallel = 25

        self.pool = ThreadPool(nodes=parallel)
        self.initialize(self.key)

        # Make sure the object was properly initialized
        assert self.shape
        assert self.chunk_shape
        assert self.dtype

    def initialize(self, path):
        cloudpath = "{}/info.txt".format(path)
        info = self.storage.get(cloudpath)
        if not info and not self.shape:
            name, dataset, version = self.key.split('/')[-3:]
            raise ArrayNotFound('Could not identify array with name {}/{}:{}. Please make sure the array name is correct.'.format(name, dataset, version))
        if info:
            info = json.loads(info.decode('utf-8'))
            self.shape = info['shape']
            self.chunk_shape = info['chunk_shape']
            self.key = info['key']
            self.dtype = info['dtype']
            self.order = info['order']
            self.protocol = info['protocol']
        else:
            info = {
                'shape': self.shape, 
                'chunk_shape': self.chunk_shape,
                'key': self.key,
                'dtype': self.dtype,
                'order': self.order,
                'protocol': self.protocol,
            }
            info = json.dumps(info)
            self.storage.put(cloudpath, info, content_type=None)

    def generate_cloudpaths(self, slices):
        # Slices -> Bbox
        slices = Bbox(Vec.zeros(self.shape), self.shape).reify_slices(slices)
        requested_bbox = Bbox.from_slices(slices)

        # Make sure chunks fit
        full_bbox = requested_bbox.expand_to_chunk_size(
            self.chunk_shape, offset = Vec.zeros(self.shape)
        )

        # Clamb the border
        full_bbox = Bbox.clamp(full_bbox, Bbox(Vec.zeros(self.shape), self.shape))

        # Generate chunknames
        cloudpaths = list(chunknames(
            full_bbox, self.shape, 
            self.key, self.chunk_shape, 
            protocol=self.protocol
        ))
        return cloudpaths, requested_bbox 

    # read from raw file and transform to numpy array 
    def decode(self, chunk):
        return np.frombuffer(bytearray(chunk), dtype=self.dtype).reshape(self.chunk_shape, order='F')

    def download_chunk(self, cloudpath):
        chunk = self.storage.get(cloudpath)
        if chunk:
            chunk = self.decode(chunk)
        else: 
            chunk = np.zeros(shape=self.chunk_shape, dtype=self.dtype, order=self.order)
        bbox = Bbox.from_filename(cloudpath)
        return chunk, bbox

    def download(self, cloudpaths, requested_bbox):
        # Download chunks
        chunks_bboxs = self.pool.map(self.download_chunk, cloudpaths)

        # Combine Chunks
        renderbuffer = np.zeros(shape=requested_bbox.to_shape(), dtype=self.dtype, order=self.order)
        def process(chunk_bbox):
            chunk, bbox = chunk_bbox
            shade(renderbuffer, requested_bbox, chunk, bbox)
        self.pool.map(process, chunks_bboxs)

        return renderbuffer

    def squeeze(self, slices, tensor):
        squeeze_dims = [] 

        if isinstance(slices, list) and len(slices)==1:
            slices = slices[0]

        if not isinstance(slices, list) and not isinstance(slices, tuple):
            slices = [slices]
            
        for dim in range(len(slices)):
            if isinstance(slices[dim], int):
                squeeze_dims.append(dim)

        if len(squeeze_dims) >= 1:
            tensor = tensor.squeeze(axis=(*squeeze_dims, ) )

        if len(tensor.shape) == 0:
            tensor = tensor.item()

        return tensor

    def __getitem__(self, slices):
        cloudpaths, requested_bbox = self.generate_cloudpaths(slices)
        tensor = self.download(cloudpaths, requested_bbox)
        tensor = self.squeeze(slices, tensor)
        return tensor

    def encode(self, chunk):
        return chunk.tostring('F')

    def upload_chunk(self, cloudpath_chunk):
        cloudpath, chunk = cloudpath_chunk
        chunk = self.encode(chunk)
        chunk = self.storage.put(cloudpath, chunk, content_type=None)
        
    def chunkify(self, cloudpaths, requested_bbox, item):
        chunks = []
        for path in cloudpaths:
            cloudchunk = Bbox.from_filename(path)
            intersection = Bbox.intersection(cloudchunk, requested_bbox)
            chunk_slices = (intersection-cloudchunk.minpt).to_slices()
            item_slices = (intersection-requested_bbox.minpt).to_slices()

            chunk = np.zeros(shape=self.chunk_shape, dtype=self.dtype, order=self.order)
            if np.any(np.array(intersection.to_shape()) != np.array(self.chunk_shape)):
                logger.debug('Non aligned write')
                chunk, _ = self.download_chunk(path)
            else:
                chunk = np.zeros(shape=self.chunk_shape, dtype=self.dtype, order=self.order)

            chunk[chunk_slices] = item[item_slices]
            chunks.append(chunk)

        return zip(cloudpaths, chunks)
    
    def upload(self, cloudpaths, requested_bbox, item):
        try:
            item = np.broadcast_to(item, requested_bbox.to_shape())
        except ValueError as err:
            raise IncompatibleBroadcasting(err)

        try:
            item = item.astype(self.dtype)
        except Exception as err:
            raise IncompatibleTypes(err)

        cloudpaths_chunks = self.chunkify(cloudpaths, requested_bbox, item)
        self.pool.map(self.upload_chunk, list(cloudpaths_chunks))

    def __setitem__(self, slices, item):
        cloudpaths, requested_bbox = self.generate_cloudpaths(slices)
        self.upload(cloudpaths, requested_bbox, item)




# Taken from https://github.com/hsjeong5/MNIST-for-Numpy
import hub
import numpy as np
from urllib import request
import gzip
import pickle
import time

filename = [
    ["train","train-images-idx3-ubyte.gz"],
    ["train_labels","t10k-images-idx3-ubyte.gz"],
    ["test","train-labels-idx1-ubyte.gz"],
    ["test_labels","t10k-labels-idx1-ubyte.gz"]
]

def download_mnist():
    base_url = "http://yann.lecun.com/exdb/mnist/"
    for name in filename:
        print("Downloading "+name[1]+"...")
        request.urlretrieve(base_url+name[1], name[1])
    print("Download complete.")

def save_mnist():
    mnist = {}
    for name in filename[:2]:
        with gzip.open(name[1], 'rb') as f:
            mnist[name[0]] = np.frombuffer(f.read(), np.uint8, offset=16).reshape(-1,28*28)
    for name in filename[-2:]:
        with gzip.open(name[1], 'rb') as f:
            mnist[name[0]] = np.frombuffer(f.read(), np.uint8, offset=8)
    with open("mnist.pkl", 'wb') as f:
        pickle.dump(mnist,f)
    print("Save complete.")

def init():
    download_mnist()
    save_mnist()

def load():
    with open("mnist.pkl",'rb') as f:
        mnist = pickle.load(f)
    return mnist

if __name__ == '__main__':
    t1 = time.time()
    #init()
    t2 = time.time()
    print('downloading takes {}s'.format(t2-t1))
    arrays = load()
    chunk_length = 128
    for key in arrays:
        t1 = time.time()
        obj = arrays[key]
        shape = obj.shape
        chunk_size = np.array(shape)
        chunk_size[0] = chunk_length
        x = hub.array(shape, name='mnist/mnist:{}'.format(key), chunk_size=chunk_size, dtype='uint8')
        x[:] = obj 
        t2 = time.time()
        print('uploaded {} {} in {}s'.format(key, shape, t2-t1))

# Download MNIST
# Unzip
# Upload
# Download



import warnings
from hub.utils.store_control import StoreControlClient
from hub.marray.interface import HubArray
import numpy as np
from hub.exceptions import WrongTypeError

def _get_path(name, public=False):
    if len(name.split('/')) == 1:
        name = '{}/{}'.format(name,name)
    user = name.split('/')[0]
    dataset = name.split('/')[1].split(':')[0]
    tag = name.split(':')
    if len(tag) == 1:
        tag.append('latest')
    tag = tag[1]
    bucket = StoreControlClient().get_config(public)['BUCKET']
    if bucket=='':
        exit()
    path = 's3://'+bucket+'/'+user+'/'+dataset+'/'+tag
    #print(path)
    return path

def load(name):
    is_public = name in ['imagenet', 'cifar', 'coco', 'mnist']
    path = _get_path(name, is_public)
    return HubArray(key=path, public=is_public)

def array(shape=None, name=None, dtype='float', chunk_size=None):

    if not name:
        raise Exception('No name provided, please name your array - hub.array(..., name="username/dataset:version") ')

    path = _get_path(name)

    if not shape:
        return load(name)

    return create(path, shape, dtype, chunk_size)
    
def create(path, dim=[50000, 28, 28], dtype='float', chunk_size=None):
    # auto chunking
    if chunk_size is None:
        chunk_size = list(dim)
        chunk_size[0] = 1

    # Input checking
    assert len(chunk_size) == len(dim)
    assert np.array(dim).dtype in np.sctypes['int']
    assert np.array(chunk_size).dtype in np.sctypes['int']
    
    try:
        np.zeros((1), dtype=dtype)
    except:
        raise WrongTypeError('Dtype {} is not supported '.format(dtype))

    return HubArray(shape=dim, dtype=dtype, chunk_shape=chunk_size, key=path, protocol=None)
    
def delete(name):
    path = _get_path(name)
    bucket = StoreControlClient().get_config()['BUCKET']
    s3.Object(bucket, path.split(bucket)[-1]).delete()



import json
import hub.config as config
from hub.log import logger
import os
from hub.utils.verify import Verify
import base64
from hub.utils.base import gen
from hub.exceptions import NotAuthorized

class StoreControlClient(object):
    """
    Controlling Snark Pods through rest api
    """
    def __init__(self):
        super(StoreControlClient, self).__init__()

    def is_authenticated(cls):
        return os.path.getsize(config.TOKEN_FILE_PATH) > 10

    @classmethod
    def lookup_aws_creds(self):
        with open(config.AWSCRED_PATH, 'r') as file:
            creds = file.readlines()
            
            for line in creds:
                line = ''.join(line.split())
                if 'aws_access_key_id' in line:
                    access_key = line.split('=')[1]
                if 'aws_secret_access_key' in line:
                    secret_key = line.split('=')[1]

        success, creds = Verify(access_key, secret_key).verify_aws(bucket=None)
        if success:
            self.save_config(creds)
        else:
            raise Exception('Error in lookup AWS credentials or in having AWS S3 bucket access.')
        return creds
            
    @classmethod
    def get_config(self, public=False):
        try:
            if os.path.exists(config.TOKEN_FILE_PATH):
                with open(config.TOKEN_FILE_PATH, 'r') as file:
                    details = file.readlines()
                    details = json.loads(''.join(details))
                    return details
            else:
                return gen()

            # Disabled AWS credential lookup
            if False and os.path.exists(config.AWSCRED_PATH):
                return self.lookup_aws_creds()
            raise Exception
        except Exception as err:
            raise NotAuthorized("No Hub or AWS credentials found. Please provide credentials by running '> hub configure'")
            return {
                'AWS_ACCESS_KEY_ID':'',
                'AWS_SECRET_ACCESS_KEY': '',
                'BUCKET': ''
            }
        return details

    @classmethod
    def save_config(self, details):
        try:
            os.mkdir('/'.join(config.TOKEN_FILE_PATH.split('/')[:-1]))
        except OSError:
            pass

        with open(config.TOKEN_FILE_PATH, 'w') as file:
            file.writelines(json.dumps(details))

    def purge_token(self):
        if os.path.isfile(config.TOKEN_FILE_PATH):
            os.remove(config.TOKEN_FILE_PATH)